/*-----------------------------------------------------------------------

File  : chatlib.c

Author: Stephan Schulz

Contents
 

  Copyright 1998-2002 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.

Changes

<1> Mon Nov 18 02:30:35 CET 2002
    New

-----------------------------------------------------------------------*/

#include "chatlib.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: SecureWrite() 
//
// As write, but calculate lenght.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

int  SecureWrite(int fd, char* message)
{
   int res, len;

   len = strlen(message);
   res = write(fd, message, len);
   /* if(res!=len)
   {
      SysErr("write");
      }*/
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: CreateListener()
// 
//  Create a non-blocking socket bound to the named port and return
//  the file descriptor.  
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

int CreateListener(int port, SockAdmin_p sock_cell)
{
   int sock;
   long sock_flags;
   
   sock = socket(PF_INET, SOCK_STREAM, 0);
   if(sock == -1)
   {
      SysErr("socket");
   }
   memset(&(sock_cell->sock_name), 0, sizeof(sock_cell->sock_name));
   (sock_cell->sock_name).sin_family = AF_INET;
   (sock_cell->sock_name).sin_port = htons(port);
   (sock_cell->sock_name).sin_addr.s_addr = htonl(INADDR_ANY);
   if (bind(sock, (struct sockaddr *) &(sock_cell->sock_name), 
	    sizeof(struct sockaddr_in)) < 0)
   {
      SysErr("bind");
   } 
   sock_flags = fcntl(sock, F_GETFL);
   fcntl(sock, F_SETFL, sock_flags|O_NONBLOCK);
   if(listen(sock, MAX_CONNECTIONS) == -1)
   {
      SysErr("listen");
   }
   sock_cell->sock_fd = sock;
   sock_cell->state = PSListening;
   printf("Server is listening on port %d with file descriptor %d\n",
	  port, sock);
   return sock;
}


/*-----------------------------------------------------------------------
//
// Function: InitializeConnectionArray()
//
//   Initialize the first size cells of a connection array.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void ConnectionArrayInitialize(SockAdmin_p array)
{
   int i;
   for(i=0; i<MAX_CONNECTIONS;i++)
   {
      array[i].sock_fd = -1;
      array[i].state = PSUnused;
   }
}


/*-----------------------------------------------------------------------
//
// Function: ConnectionArrayFind()
//
//   Find the smallest index such that array.state = state (-1, if no
//   such index exists).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

int  ConnectionArrayFind(SockAdmin_p array, ProtocolState state)
{
   int i;

   for(i=0; i<MAX_CONNECTIONS; i++)
   {
      if(array[i].state == state)
      {
	 return i;
      }
   }
   return -1;
}


/*-----------------------------------------------------------------------
//
// Function: SelectSetInit()
//
//   Mark all sockets in array that are not in state PSUnused in
//   set. Return largest file descriptor + 1.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

int SelectSetInit(SockAdmin_p array, fd_set *set)
{
   int i, max=0;

   FD_ZERO(set);
   for(i=0; i<MAX_CONNECTIONS; i++)
   {
      if(array[i].state != PSUnused)
      {	 
	 FD_SET(array[i].sock_fd, set);
	 max = max<array[i].sock_fd?array[i].sock_fd:max; 
      }
   }
   return max+1;
}


/*-----------------------------------------------------------------------
//
// Function: CreateNewConnection()
//
//   Create a new connection and greet the user. Returns index of new
//   connection in connections.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

int CreateNewConnection(SockAdmin_p connections, int l_index)
{
   int    c_index;
   int    sock;
   int    sock_len;
   struct sockaddr_in err_sock_name;
   
   assert(connections[l_index].state = PSListening);

   c_index = ConnectionArrayFindUnused(connections);
   if(c_index == -1)
   { /* No space for new connection */
      sock_len = sizeof(err_sock_name);   
      sock = accept(connections[l_index].sock_fd, 
		    (struct sockaddr*) &err_sock_name, 
		    &sock_len);
      if(sock != -1)
      {
	 SecureWrite(sock, "All connections in use, try again later.\n");
	 close(sock);
      }
      /* Silently ignore else case - may be network error, or
	 prematurely closed connection, or... */
   }
   else
   {
      sock_len = sizeof(connections[c_index].sock_name);
      sock = accept(connections[l_index].sock_fd,
		    (struct sockaddr*) &(connections[c_index].sock_name),
		    &sock_len);      
      if(sock != -1)
      {
	 connections[c_index].sock_fd = sock;
	 connections[c_index].state = PSConnected;
	 SecureWrite(sock, "Successfully connected.\n");
      }
      /* Silently ignore else case - may be network error, or
	 prematurely closed connection, or... */
   }
   printf("Connection established on index %d.\n", c_index);   
   return c_index;
}
	

/*-----------------------------------------------------------------------
//
// Function: HandleMessage()
//
//   Read a message and print it to all connected users except the
//   source. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void HandleMessage(SockAdmin_p connections, int c_index)
{
   char buff[DEFAULT_MSG_LEN];
   ssize_t msglen;
   int i;

   msglen = read(connections[c_index].sock_fd, buff, DEFAULT_MSG_LEN);
   if(msglen == -1)
   { /* Ignore errors */
      return;
   }
   if(msglen == 0)
   {
      SecureWrite(connections[c_index].sock_fd, "Ok, bye bye.\n");
      close(connections[c_index].sock_fd);
      connections[c_index].state = PSUnused;
      connections[c_index].sock_fd = -1;
      printf("Connection on %d terminated.\n", c_index);
      return;
   }
   for(i=0; i< MAX_CONNECTIONS; i++)
   {
      if(connections[i].state == PSConnected && i!=c_index)
      {
	 write(connections[i].sock_fd, buff,msglen);
      }
   }
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/










